#!/usr/bin/env python3

import argparse
import datetime
import os
import pickle
import urllib.parse
from typing import Generator, Optional, Tuple

import graphviz
import requests

from atlassian import Confluence


class DataModel:

    DEBUG = False

    def __init__(self,
                 url: str,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 user_id: int = 0,
                 cacert: str = '') -> None:
        self.objects_url = urllib.parse.urljoin(url, "/v1/objectType")
        self.facts_url = urllib.parse.urljoin(url, "/v1/factType")
        self.username = username
        self.password = password
        self.user_id = user_id
        self.status: Optional[int] = None
        self._objects: Optional[dict] = None
        self._facts: Optional[dict] = None
        self.verify = cacert if cacert != '' else True

    def load(self) -> None:
        """Download the datamodel from the act instance"""

        headers = {
            'ACT-User-ID': str(self.user_id),
            'Accept': 'application/json'
        }
        auth = (self.username, self.password)

        if self.username:
            r = requests.get(self.objects_url, auth=auth, headers=headers, verify=self.verify)
        else:
            r = requests.get(self.objects_url, headers=headers, verify=self.verify)

        if r.status_code == 200:
            self._objects = r.json()
        else:
            if self.DEBUG:
                print("Error loading objects: {}".format(r.status_code))
            self._objects = None
            self._facts = None
            self.status = r.status_code
            return

        if self.username:
            r = requests.get(self.facts_url, auth=auth, headers=headers, verify=self.verify)
        else:
            r = requests.get(self.facts_url, headers=headers, verify=self.verify)

        if r.status_code == 200:
            self._facts = r.json()
        else:
            if self.DEBUG:
                print("Error loading objects: {}".format(r.status_code))
            self._objects = None
            self._facts = None
            self.status = r.status_code
            return

        self.status = r.status_code

    def __eq__(self, other: object) -> bool:
        return set(self.objects) == set(other.objects) and set(self.facts) == set(other.facts)

    @property
    def facts(self) -> Generator[Tuple[str, str, str, bool], None, None]:
        """Iterate over fact bindings"""

        if not self._facts:
            if self.DEBUG:
                print("Trying to iterate over None facts")
            return
        for fact in self._facts['data']:
            if not fact:
                print("Nonetype", fact)
                continue
            bindings = fact['relevantObjectBindings']
            if not bindings:
                continue
            for binding in bindings:
                yield (fact['name'],
                       binding['sourceObjectType']['name'],
                       binding['destinationObjectType']['name'] if binding['destinationObjectType'] else None,
                       binding['bidirectionalBinding'])

    @property
    def objects(self) -> Generator[str, None, None]:
        """Iterate over object bindings"""

        if not self._objects:
            if self.DEBUG:
                print("Trying to iterate over None objects")
            return

        for obj in self._objects['data']:
            if not obj:
                print("Nonetype", obj)
                continue
            yield obj['name']


def parse_args() -> argparse.Namespace:
    """Handle command line arguments, returning the arguments ns"""

    parser = argparse.ArgumentParser(description="Build a graph of the act datamodel")
    parser.add_argument('url', type=str, help="Url of the act instance to graph")
    parser.add_argument('--uid', default=1, type=int, help="Act user ID")
    parser.add_argument('--http_username', type=str, default=None, help="HTTP Username")
    parser.add_argument('--http_password', type=str, default=None, help="HTTP Password")
    parser.add_argument('--parent_id', type=int, default=None, help="Confluence upload")
    parser.add_argument('--confluence_url', type=str, default=None, help="Confluence api url")
    parser.add_argument('--confluence_user', type=str, default=None, help="Confluence user")
    parser.add_argument('--confluence_password', type=str, default=None, help="Confluence password")
    parser.add_argument('--dump_source', type=str, default=None, help="Dump graphviz source to directory")
    parser.add_argument('--cacert', type=str, default=None, help="If you need to specify a CA certificate file")

    return parser.parse_args()


def run() -> None:
    """Main program loop"""

    args = parse_args()

    dm = DataModel(args.url, args.http_username, args.http_password, args.uid, args.cacert)
    dm.load()

    if dm.status != 200:
        print("{} Status code {}".format(str(datetime.datetime.now()), dm.status))
        return

    try:
        old_dm = pickle.load(open('cache.dat', 'rb'))
        if old_dm == dm:
            return
    except FileNotFoundError:
        print("First run")

    print("{} Graphing changes".format(str(datetime.datetime.now())))

    pickle.dump(dm, open('cache.dat', 'wb'))

    dot = graphviz.Digraph(comment='Double edge facts')

    # for obj in dm.objects:
    #     dot.node(obj, obj)

    for fact in dm.facts:
        name, s, d, di = fact
        if name == 'mentions':
            continue
        if not d:
            continue

        dot.node(s, s)
        dot.node(d, d)
        if di:
            dot.edge(s, d, label=name, dir="both")
        else:
            dot.edge(s, d, label=name)

    if args.dump_source:
        with open(os.path.join(args.dump_source, 'double.dot'), 'w') as f:
            f.write(dot.source)

    dot.render('output/double', format='png', renderer='cairo')

    dot = graphviz.Digraph(comment='Single edge facts')

    for fact in dm.facts:

        name, s, d, di = fact

        if d:
            continue

        dot.node(name, label=name, shape='diamond')
        dot.node(s, s)

        dot.edge(s, name)

    if args.dump_source:
        with open(os.path.join(args.dump_source, 'single.dot'), 'w') as f:
            f.write(dot.source)

    dot.render('output/single', format='png', renderer='cairo')

    dot = graphviz.Digraph(comment='All Double edge facts')

    for obj in dm.objects:
        dot.node(obj, obj)

    for fact in dm.facts:
        name, s, d, di = fact
        if not d:
            continue

        dot.node(s, s)
        dot.node(d, d)
        if di:
            dot.edge(s, d, label=name, dir="both")
        else:
            dot.edge(s, d, label=name)

    if args.dump_source:
        with open(os.path.join(args.dump_source, 'complete.dot'), 'w') as f:
            f.write(dot.source)

    dot.render('output/complete', format='png', renderer='cairo')

    if args.parent_id:
        try:
            os.environ.pop('https_proxy')
        except KeyError:
            pass
        try:
            os.environ.pop('http_proxy')
        except KeyError:
            pass

        verify_ssl = False if args.cacert else True

        confluence = Confluence(url=args.confluence_url,
                                username=args.confluence_user,
                                password=args.confluence_password,
                                verify_ssl=verify_ssl)
        confluence.attach_file(
            'output/double.cairo.png',
            page_id=args.parent_id,
            title='Double Edged Facts')
        confluence.attach_file(
            'output/single.cairo.png',
            page_id=args.parent_id,
            title='Single Edged Facts')
        confluence.attach_file(
            'output/complete.cairo.png',
            page_id=args.parent_id,
            title='Single Edged Facts')

        if args.dump_source:
            confluence.attach_file(
                os.path.join(args.dump_source, 'complete.dot'),
                page_id=args.parent_id,
                title='complete source')
            confluence.attach_file(
                os.path.join(args.dump_source, 'double.dot'),
                page_id=args.parent_id,
                title='double source')
            confluence.attach_file(
                os.path.join(args.dump_source, 'single.dot'),
                page_id=args.parent_id,
                title='single source')



if __name__ == '__main__':
    run()



#!/usr/bin/env python3

""" Add ACT types """

import argparse
import json
import os
import sys
from logging import critical, warning
from typing import Any, Dict, List, Text

from pkg_resources import resource_filename

import act.api


def parseargs() -> argparse.Namespace:
    """ Parse arguments """
    parser = argparse.ArgumentParser(description="ACT Bootstrap data model")
    parser.add_argument(
        "action",
        nargs=1,
        choices=['add', 'print'],
        help="Action (add or print)")
    parser.add_argument(
        "--default-object-types",
        action="store_true",
        help="Default object types")
    parser.add_argument(
        "--default-fact-types",
        action="store_true",
        help="Default fact types")
    parser.add_argument(
        "--default-meta-fact-types",
        action="store_true",
        help="Default meta fact types")
    parser.add_argument(
        "--object-types-file",
        help="Object type definitions (json)")
    parser.add_argument(
        "--fact-types-file",
        help="Fact type definitions (json)")
    parser.add_argument(
        "--meta-fact-types-file",
        help="Meta Fact type definitions (json)")
    parser.add_argument('--http-user', dest='http_user', help="ACT HTTP Basic Auth user")
    parser.add_argument(
        '--http-password',
        dest='http_password',
        help="ACT HTTP Basic Auth password")
    parser.add_argument("--logfile", help="Log to file (default = stdout)")
    parser.add_argument("--loglevel", default="info", help="Loglevel (default = info)")
    parser.add_argument("--userid", type=int, help="User ID")
    parser.add_argument(
        "--act-baseurl",
        dest="act_baseurl",
        help="API URI")

    return parser.parse_args()


def etc_file(filename: Text) -> Text:
    "Get content of file from resource/etc"
    return resource_filename("act.types", "etc/{}".format(filename))  # ).decode('utf-8')


def load_types(filename: Text) -> List[Dict]:
    """
    Parse as json, and exit on parse error
    """
    if not os.path.isfile(filename):
        critical("File not found: {}".format(filename))
        sys.exit(1)

    try:
        types = json.loads(open(filename, encoding="UTF8").read())
    except json.decoder.JSONDecodeError:
        critical("Unable to parse file as json: %s" % filename)
        sys.exit(2)

    if not isinstance(types, list):
        critical("Types must be a list of dicts")
        sys.exit(3)

    return types


def print_json(o: Any) -> None:
    " Print dict as sorted, indented json object "
    print(json.dumps(o, indent=4, sort_keys=True))


def create_object_types(client: act.api.Act, object_types: List[Dict]) -> None:
    """
    Create object types
    """

    existing_object_types = [object_type.name
                             for object_type in client.get_object_types()]

    # Create all objects
    for object_type in object_types:
        name = object_type["name"]
        validator = object_type.get("validator", act.api.DEFAULT_VALIDATOR)

        if name in existing_object_types:
            warning("Object type %s already exists" % name)
            continue

        client.object_type(name=name, validator_parameter=validator).add()


def create_fact_types(client: act.api.Act, fact_types: List[Dict]) -> None:
    """
    Create fact type with allowed bindings to ALL objects
    We want to change this later, but keep it like this to make it simpler
    when evaluating the data model
    """

    for fact_type in fact_types:
        name = fact_type["name"]
        validator = fact_type.get("validator", act.api.DEFAULT_VALIDATOR)
        object_bindings = fact_type.get("objectBindings", [])

        if not object_bindings:
            client.create_fact_type_all_bindings(
                name, validator_parameter=validator)

        else:
            client.create_fact_type(name, validator=validator, object_bindings=object_bindings)


def create_meta_fact_types(client: act.api.Act, meta_fact_types: List[Dict]) -> None:
    """
    Create fact type with allowed bindings to ALL objects
    We want to change this later, but keep it like this to make it simpler
    when evaluating the data model
    """

    for meta_fact_type in meta_fact_types:
        name = meta_fact_type["name"]
        validator = meta_fact_type.get("validator", act.api.DEFAULT_VALIDATOR)
        fact_bindings = meta_fact_type.get("factBindings", [])

        if not fact_bindings:
            client.create_meta_fact_type_all_bindings(name, validator_parameter=validator)

        else:
            client.create_meta_fact_type(name, fact_bindings=fact_bindings, validator=validator)


def main() -> None:
    "Main function"
    args = parseargs()

    if "print" in args.action:
        if args.default_object_types:
            print_json(load_types(etc_file("object-types.json")))

        if args.default_fact_types:
            print_json(load_types(etc_file("fact-types.json")))

        if args.default_meta_fact_types:
            print_json(load_types(etc_file("meta-fact-types.json")))

        if args.object_types_file:
            print_json(load_types(args.object_types_file))

        if args.fact_types_file:
            print_json(load_types(args.fact_types_file))

        if args.meta_fact_types_file:
            print_json(load_types(args.meta_fact_types_file))

        sys.exit(0)

    elif "add" in args.action:

        if not (args.act_baseurl and args.userid):
            sys.stderr.write("Missing --act-baseurl and/or --userid\n")
            sys.exit(1)

        auth = None
        if args.http_user:
            auth = (args.http_user, args.http_password)

        client = act.api.Act(
            args.act_baseurl,
            args.userid,
            args.loglevel,
            args.logfile,
            "act-types",
            requests_common_kwargs={'auth': auth})

        if args.default_object_types:
            create_object_types(client, load_types(etc_file("object-types.json")))

        if args.default_fact_types:
            create_fact_types(client, load_types(etc_file("fact-types.json")))

        if args.default_meta_fact_types:
            create_meta_fact_types(client, load_types(etc_file("meta-fact-types.json")))

        if args.object_types_file:
            create_object_types(client, load_types(args.object_types_file))

        if args.fact_types_file:
            create_fact_types(client, load_types(args.fact_types_file))

        if args.meta_fact_types_file:
            create_meta_fact_types(client, load_types(args.meta_fact_types_file))


if __name__ == "__main__":
    main()
